import os
import unittest

from pythongettext.msgfmt import Msgfmt
from pythongettext.msgfmt import PoSyntaxError


FOLDER = os.path.dirname(__file__)


class TestWriter(unittest.TestCase):

    def compare_po_mo(self, poname, moname):
        po_file = None
        mo_file = None
        try:
            po_file = open(os.path.join(FOLDER, poname), 'rb')
            po = Msgfmt(po_file).get()
            mo_file = open(os.path.join(FOLDER, moname), 'rb')
            mo = b''.join(mo_file.readlines())
        finally:
            if po_file is not None:
                po_file.close()
            if mo_file is not None:
                mo_file.close()

        self.assertEqual(mo, po)

    def test_empty(self):
        self.compare_po_mo('test_empty.po', 'test_empty.mo')

    def test_test(self):
        self.compare_po_mo('test.po', 'test.mo')

    def test_test2(self):
        self.compare_po_mo('test2.po', 'test2.mo')

    def test_msgctxt(self):
        self.compare_po_mo('test3.po', 'test3.mo')

    def test_test4(self):
        po_file = open(os.path.join(FOLDER, 'test4.po'), 'rb')
        po = Msgfmt(po_file)
        po.read(header_only=True)
        po_file.close()
        self.assertTrue(
            po.messages[''].startswith('Project-Id-Version: foo'))
        self.assertEqual(po.encoding, 'iso-8859-1')

    def test_test5(self):
        po_file = open(os.path.join(FOLDER, 'test5.po'), 'rb')
        po = Msgfmt(po_file)
        try:
            with self.assertRaises(PoSyntaxError):
                po.read()
        finally:
            po_file.close()
        self.assertEqual(po.encoding, 'utf-8')

    def test_test5_unicode_name(self):
        po_file = open(os.path.join(FOLDER, 'test5.po'), 'rb')
        po = Msgfmt(po_file, name='dømain')
        try:
            with self.assertRaises(PoSyntaxError):
                po.read()
        finally:
            po_file.close()
        self.assertEqual(po.encoding, 'utf-8')

    def test_test6(self):
        self.compare_po_mo('test6.po', 'test6.mo')

    def test_test6_unicode_header(self):
        po_file = open(os.path.join(FOLDER, 'test6.po'), 'rb')
        po = Msgfmt(po_file)
        po.read(header_only=True)
        po_file.close()
        self.assertTrue(po.messages[''].startswith(
            'Project-Id-Version: Tøst 1.0'))
        self.assertEqual(po.encoding, 'utf-8')

    def test_escape(self):
        po_file = open(os.path.join(FOLDER, 'test_escape.po'), 'rb')
        po = Msgfmt(po_file)
        try:
            with self.assertRaises(PoSyntaxError) as e:
                po.read()
            self.assertTrue('line 19' in e.exception.msg)
            self.assertEqual(po.encoding, 'utf-8')
        finally:
            po_file.close()

    def test_unicode_bom(self):
        self.compare_po_mo('test_unicode_bom.po', 'test_unicode_bom.mo')

    def test_plural(self):
        po_file = open(os.path.join(FOLDER, 'test_plural.po'), 'rb')
        po = Msgfmt(po_file)
        try:
            po.read()
        finally:
            po_file.close()
        self.assertEqual(
            set(po.messages.keys()),
            {'', 'm1', 'm2 ø\x00{d} ømsgid',
                 'øcontext\x04m3 ø\x00{d} ømsgid context'})



#! /usr/bin/env python
# Written by Martin v. Loewis <loewis@informatik.hu-berlin.de>
#
# Changed by Christian 'Tiran' Heimes <tiran@cheimes.de> for the placeless
# translation service (PTS) of Zope
#
# Fixed some bugs and updated to support msgctxt
# by Hanno Schlichting <hanno@hannosch.eu>

"""Generate binary message catalog from textual translation description.

This program converts a textual Uniforum-style message catalog (.po file) into
a binary GNU catalog (.mo file). This is essentially the same function as the
GNU msgfmt program, however, it is a simpler implementation.

This file was taken from Python-2.3.2/Tools/i18n and altered in several ways.
Now you can simply use it from another python module:

  from msgfmt import Msgfmt
  mo = Msgfmt(po).get()

where po is path to a po file as string, an opened po file ready for reading or
a list of strings (readlines of a po file) and mo is the compiled mo file as
binary string.

Exceptions:

  * IOError if the file couldn't be read

  * msgfmt.PoSyntaxError if the po file has syntax errors
"""

import array
import codecs
import io
import struct
from ast import literal_eval
from email.parser import HeaderParser


def header_charset(s):
    p = HeaderParser()
    return p.parsestr(s).get_content_charset()


class PoSyntaxError(Exception):
    """ Syntax error in a po file """

    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return 'Po file syntax error: %s' % self.msg


class Msgfmt:

    def __init__(self, po, name='unknown'):
        self.po = po
        self.name = name
        self.messages = {}
        self.openfile = False
        # Start off assuming latin-1, so everything decodes without failure,
        # until we know the exact encoding
        self.encoding = 'latin-1'

    def readPoData(self):
        """ read po data from self.po and return an iterator """
        output = []
        if isinstance(self.po, str):
            output = open(self.po, 'rb')
        elif isinstance(self.po, io.IOBase):
            self.po.seek(0)
            self.openfile = True
            output = self.po
        elif isinstance(self.po, list):
            output = self.po
        if not output:
            raise ValueError("self.po is invalid! %s" % type(self.po))
        if isinstance(output, io.IOBase):
            # remove BOM from the start of the parsed input
            first = output.readline()
            if len(first) == 0:
                return output.readlines()
            if first.startswith(codecs.BOM_UTF8):
                first = first.lstrip(codecs.BOM_UTF8)
            return [first] + output.readlines()
        return output

    def add(self, context, id, string, fuzzy):
        "Add a non-empty and non-fuzzy translation to the dictionary."
        if string and not fuzzy:
            # The context is put before the id and separated by a EOT char.
            if context:
                id = context + '\x04' + id
            if not id:
                # See whether there is an encoding declaration
                charset = header_charset(string)
                if charset:
                    # decode header in proper encoding
                    string = string.encode(self.encoding).decode(charset)
                    self.encoding = charset
            self.messages[id] = string

    def generate(self):
        "Return the generated output."
        # the keys are sorted in the .mo file
        keys = sorted(self.messages.keys())
        offsets = []
        ids = strs = b''
        for id in keys:
            msg = self.messages[id].encode(self.encoding)
            id = id.encode(self.encoding)
            # For each string, we need size and file offset. Each string is
            # NUL terminated; the NUL does not count into the size.
            offsets.append((len(ids), len(id), len(strs),
                            len(msg)))
            ids += id + b'\0'
            strs += msg + b'\0'
        output = b''
        # The header is 7 32-bit unsigned integers. We don't use hash tables,
        # so the keys start right after the index tables.
        keystart = 7 * 4 + 16 * len(keys)
        # and the values start after the keys
        valuestart = keystart + len(ids)
        koffsets = []
        voffsets = []
        # The string table first has the list of keys, then the list of values.
        # Each entry has first the size of the string, then the file offset.
        for o1, l1, o2, l2 in offsets:
            koffsets += [l1, o1 + keystart]
            voffsets += [l2, o2 + valuestart]
        offsets = koffsets + voffsets
        # Even though we don't use a hashtable, we still set its offset to be
        # binary compatible with the gnu gettext format produced by:
        # msgfmt file.po --no-hash
        output = struct.pack("Iiiiiii",
                             0x950412de,        # Magic
                             0,                 # Version
                             len(keys),         # # of entries
                             7 * 4,             # start of key index
                             7 * 4 + len(keys) * 8,  # start of value index
                             0, keystart)       # size and offset of hash table
        output += array.array("i", offsets).tobytes()
        output += ids
        output += strs
        return output

    def get(self):
        """ """
        self.read()
        # Compute output
        return self.generate()

    def read(self, header_only=False):
        """ """
        ID = 1
        STR = 2
        CTXT = 3

        section = None
        fuzzy = 0
        msgid = msgstr = msgctxt = ''

        # Parse the catalog
        lno = 0
        for l_ in self.readPoData():
            l_ = l_.decode(self.encoding)
            lno += 1
            # If we get a comment line after a msgstr or a line starting with
            # msgid or msgctxt, this is a new entry
            if section == STR and (l_[0] == '#' or (l_[0] == 'm' and
               (l_.startswith('msgctxt') or l_.startswith('msgid')))):
                self.add(msgctxt, msgid, msgstr, fuzzy)
                section = None
                fuzzy = 0
                # If we only want the header we stop after the first message
                if header_only:
                    break
            # Record a fuzzy mark
            if l_[:2] == '#,' and 'fuzzy' in l_:
                fuzzy = 1
            # Skip comments
            if l_[0] == '#':
                continue
            # Now we are in a msgctxt section
            if l_.startswith('msgctxt'):
                section = CTXT
                l_ = l_[7:]
                msgctxt = ''
            # Now we are in a msgid section, output previous section
            elif (l_.startswith('msgid') and
                  not l_.startswith('msgid_plural')):
                if section == STR:
                    self.add(msgid, msgstr, fuzzy)
                section = ID
                l_ = l_[5:]
                msgid = msgstr = ''
                is_plural = False
            # This is a message with plural forms
            elif l_.startswith('msgid_plural'):
                if section != ID:
                    raise PoSyntaxError(
                        'msgid_plural not preceeded by '
                        'msgid on line %d of po file %s' %
                        (lno, repr(self.name)))
                l_ = l_[12:]
                msgid += '\0'  # separator of singular and plural
                is_plural = True
            # Now we are in a msgstr section
            elif l_.startswith('msgstr'):
                section = STR
                if l_.startswith('msgstr['):
                    if not is_plural:
                        raise PoSyntaxError(
                            'plural without msgid_plural '
                            'on line %d of po file %s' %
                            (lno, repr(self.name)))
                    l_ = l_.split(']', 1)[1]
                    if msgstr:
                        # Separator of the various plural forms
                        msgstr += '\0'
                else:
                    if is_plural:
                        raise PoSyntaxError(
                            'indexed msgstr required for '
                            'plural on line %d of po file %s' %
                            (lno, repr(self.name)))
                    l_ = l_[6:]
            # Skip empty lines
            l_ = l_.strip()
            if not l_:
                continue
            # TODO: Does this always follow Python escape semantics?
            try:
                l_ = literal_eval(l_)
            except Exception as msg:
                raise PoSyntaxError(
                    '%s (line %d of po file %s): \n%s' %
                    (msg, lno, repr(self.name), l_))
            if isinstance(l_, bytes):
                l_ = l_.decode(self.encoding)
            if section == CTXT:
                msgctxt += l_
            elif section == ID:
                msgid += l_
            elif section == STR:
                msgstr += l_
            else:
                raise PoSyntaxError(
                    'error on line %d of po file %s' %
                    (lno, repr(self.name)))

        # Add last entry
        if section == STR:
            self.add(msgctxt, msgid, msgstr, fuzzy)

        if self.openfile:
            self.po.close()

    def getAsFile(self):
        return io.BytesIO(self.get())



import os.path

from setuptools import find_packages
from setuptools import setup


version = '5.0'

install_requires = []

here = os.path.abspath(os.path.dirname(__file__))
try:
    README = open(os.path.join(here, 'README.rst')).read()
    CHANGES = open(os.path.join(here, 'CHANGES.rst')).read()
except OSError:
    README = CHANGES = ''

setup(
    name='python-gettext',
    version=version,
    description="Python Gettext po to mo file compiler.",
    long_description=README + '\n\n' + CHANGES,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Internationalization',
        'Topic :: Software Development :: Localization',
    ],
    keywords='Python Gettext Msgctxt',
    author='Hanno Schlichting',
    author_email='hanno@hannosch.eu',
    url='https://github.com/hannosch/python-gettext',
    license='BSD',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    python_requires='>=3.7',
    install_requires=install_requires,
    include_package_data=True,
    zip_safe=False,
)
